
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import control as ct
import struct
from queue import Queue
from math import ceil
import json

from scipy.signal import cont2discrete
from .thermalsys import ThermalSystemIoT, PATH_DATA, PATH_DEFAULT

def float2hex(value):
    val_binary = struct.pack('>f', value)
    return val_binary.hex()


def long2hex(value):
    val_binary = struct.pack('>L', value)
    return val_binary.hex()


def hex2float(str):
    value = struct.unpack('!f', bytes.fromhex(str))[0]
    return value


def hex2long(str):
    value = struct.unpack('!L', bytes.fromhex(str))[0]
    return value


def signal2hex(signal):
    hstring = ""
    for point in signal:
        hstring += float2hex(point)
    return hstring


def matrix2hex(matrix):
    hstring = ""
    for row in matrix:
        for element in row:
            hstring += float2hex(element)
    return hstring


def time2hex(time_points):
    hstring = ""
    for t in time_points:
        hstring += long2hex(t)
    return hstring

def set_reference(system, ref_value=50):
    ref_hex = float2hex(ref_value)
    topic_pub = system.codes["USER_SYS_SET_REF"]
    message = json.dumps({"reference": ref_hex})
    system.connect()
    system.publish(topic_pub, message)
    system.disconnect()
    rcode = True
    print("succesfull change of reference")
    return rcode


def set_pid(system, kp=1, ki=0.4, kd=0, N=5, beta=1):
    topic_pub = system.codes["USER_SYS_SET_PID"]
    kp_hex = float2hex(kp)
    ki_hex = float2hex(ki)
    kd_hex = float2hex(kd)
    N_hex = float2hex(N)
    beta_hex = float2hex(beta)
    message = json.dumps({"kp": kp_hex,
                          "ki": ki_hex,
                          "kd": kd_hex,
                          "N": N_hex,
                          "beta": beta_hex})
    system.connect()
    system.publish(topic_pub, message)
    system.disconnect()
    rcode = True
    print("succesfull change of PID parameters")
    return rcode


def step_closed(system, r0=0 , r1=100, t0=0 ,  t1=1):
    def step_message(system, userdata, message):
        q.put(message)



    low_val = r0
    high_val = r1
    low_time = t0
    high_time = t1
    topic_pub = system.codes["USER_SYS_STEP_CLOSED"]
    topic_sub = system.codes["SYS_USER_SIGNALS_CLOSED"]
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]
    points_high = round(high_time / sampling_time) 
    points_low = round(low_time / sampling_time)
    points = points_low + points_high
    points_low_hex = long2hex(points_low)
    points_high_hex = long2hex(points_high)
    low_val_hex = float2hex(low_val)
    high_val_hex = float2hex(high_val)
    message = json.dumps({"low_val": low_val_hex,
                          "high_val": high_val_hex,
                          "points_low": points_low_hex,
                          "points_high": points_high_hex,
                          })

    system.client.on_message = step_message
    system.connect()
    system.subscribe(topic_sub)
    system.publish(topic_pub, message)
    q = Queue()
    y = []
    r = []
    u = []
    t = []


    # Setting the graphics configuration for visualizing the experiment
    fig, (ay, au) = plt.subplots(nrows=2, ncols=1, width_ratios = [1], height_ratios= [4,1], figsize=(16, 9))
    fig.set_facecolor('#b7c4c8f0')

    # settings for the upper axes, depicting the model and speed data
    ay.set_title(f'Closed loop step response experiment with an initial value of  $r_0=${r0:0.2f} and a  final value of $r_0=${r1:0.2f}')
    ay.set_ylabel('')
    ay.grid(True);
    ay.grid(color='#806600ff', linestyle='--', linewidth=0.25)
    ay.set_facecolor('#f4eed7ff')
    ay.set_xlim(0, t0 + t1  - sampling_time)

    #Setting the limits of figure
    py = 0.3
    delta_r = abs(r1 - r0)
    ylimits = [r0 , r1]
    ylimits = [np.min(ylimits)- py * delta_r , np.max(ylimits) + py * delta_r]

    ay.set_ylim(ylimits[0], ylimits[1])

    au.set_facecolor('#d7f4e3ff')
    au.set_ylim(0, 100)
    au.set_xlim(0, t0 + t1 - sampling_time )
    au.grid(color='#008066ff', linestyle='--', linewidth=0.25)


    line_r, = ay.plot(t, r, drawstyle='steps-post', color="#008066ff", linewidth=1.25)
    line_y, = ay.plot(t, y, color="#ff0066ff")
    line_u, = au.plot(t, u, color="#0066ffff")





    box = dict(boxstyle='round,pad=0.5', facecolor='white', edgecolor='white', alpha=0.75)
    # y_txt = ay.text( t0 + t1- 3, ylimits[0] + 1 , 'Temperature:\n ', fontsize=15, color="#ff6680",ha='right', va='bottom', bbox=box)
    # u_txt = au.text(t0 + t1 - 3 , 15, f'Input:', fontsize=15, color="#0066ffB0",ha='right', va='bottom', bbox=box)


    exp = []
    n = -1
    sync = False

    while n < points:
        try:
            message = q.get(True, 20 * sampling_time)
        except:
            raise TimeoutError("The connection has been lost. Please try again")

        decoded_message = str(message.payload.decode("utf-8"))
        msg_dict = json.loads(decoded_message)
        n_hex = str(msg_dict["np"])
        n = hex2long(n_hex)
        if n == 0:
            sync = True
        if sync == True:
            t_curr = n * sampling_time
            t.append(t_curr)
            y_curr = hex2float(msg_dict["y"])
            y.append(y_curr)
            r_curr = hex2float(msg_dict["r"])
            r.append(r_curr)
            u_curr = hex2float(msg_dict["u"])
            u.append(u_curr)
            line_r.set_data(t, r)
            line_y.set_data(t, y)
            line_u.set_data(t, u)
            ay.legend([line_r, line_y], [f'$r(t):$ {r_curr:0.2f}', f'$y(t):$ {y_curr: 0.3f}$~^oC$'], fontsize=16, loc="upper left")
            au.legend([line_u], [f'$u(t):$ {u_curr: 0.1f}'], fontsize=16)
            exp.append([t_curr, r_curr, y_curr, u_curr])
            plt.draw()
            plt.pause(sampling_time)


    np.savetxt(PATH_DEFAULT + "Thermal_step_closed_exp.csv",  exp, delimiter=",", fmt="%0.8f", comments="", header='t,r,y,u')
    np.savetxt(PATH_DATA + "Thermal_step_closed_exp.csv",  exp, delimiter=",", fmt="%0.8f", comments="", header='t,r,y,u')
    system.disconnect()
    plt.show()
    return t, r, y, u


def stairs_closed(system, stairs=[40, 50, 60], duration=100):
    def usersignal_message(system, userdata, message):
        q.put(message)

    # accessing fields of system IoT
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]
    topic_pub = system.codes["USER_SYS_STAIRS_CLOSED"]
    topic_sub = system.codes["SYS_USER_SIGNALS_CLOSED"]
    points = len(stairs)
    points_hex = long2hex(points)
    signal_hex = signal2hex(stairs)
    duration_points = math.ceil(duration / sampling_time)
    duration_points_hex = long2hex(duration_points)
    message = json.dumps({"signal": signal_hex, "duration": duration_points_hex, "points": points_hex})
    system.client.on_message = usersignal_message
    system.connect()
    system.subscribe(topic_sub)
    system.publish(topic_pub, message)
    q = Queue()
    np = -1
    y = []
    r = []
    t = []
    u = []
    exp = []
    fig, ax = plt.subplots()
    line_r, = ax.plot(t, r, 'b-')
    line_y, = ax.plot(t, y, 'r-')
    ax.set_xlim(0, sampling_time * duration_points * points)
    ax.set_ylim(20, 100)
    plt.grid()
    npc = 1
    while npc <= duration_points * points - 1:
        try:
            message = q.get(True, 20 * sampling_time)
        except:
            raise TimeoutError("The connection has been lost. Please try again")

        decoded_message = str(message.payload.decode("utf-8"))
        msg_dict = json.loads(decoded_message)
        np_hex = str(msg_dict["np"])
        np = hex2long(np_hex)
        print(np, hex2float(msg_dict["r"]), hex2float(msg_dict["y"]))
        if np == npc:
            t_curr = (np - 1) * sampling_time
            t.append(t_curr)
            y_curr = hex2float(msg_dict["y"])
            y.append(y_curr)
            r_curr = hex2float(msg_dict["r"])
            r.append(r_curr)
            u_curr = hex2float(msg_dict["u"])
            u.append(u_curr)
            line_r.set_data(t, r)
            line_y.set_data(t, y)
            exp.append([t_curr, r_curr, y_curr, u_curr])
            plt.draw()
            plt.pause(0.1)
            npc += 1


    npy.savetxt(PATH_DEFAULT + "Thermal_stairs_closed_exp.csv", exp, delimiter=",", fmt="%0.8f", comments="", header='t,r,y,u')
    npy.savetxt(PATH_DATA + "Thermal_stairs_closed_exp.csv", exp, delimiter=",", fmt="%0.8f", comments="", header='t,r,y,u')
    system.disconnect()
    plt.show()
    return t, y, r, u



def set_controller(system, controller):


    topic_pub = system.codes["USER_SYS_SET_GENCON"]
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]
    struct = len(controller.den[0])
    type_control = struct - 1


    if struct == 1:
        con = ct.tf(ct.tf(controller.num[0][0], controller.den[0][0]))
        N1, D1 = ct.tfdata(con)
        N1 = N1[0][0]
        D1 = D1[0][0]
        N1 = N1 / D1[0]
        D1 = D1 / D1[0]

        if len(N1) == len(D1):
            d1 = N1[0]
            N1 = N1 - d1* D1
            N1 = N1[1:]
        else:
            d1 = 0

        DB = np.array([-D1[1:]])
        DB = DB.T
        size = len(D1)-2
        In_1 = np.eye(size)
        ZR = np.zeros((1,size))
        Acon = np.block([[In_1], [ZR]])
        Acon = np.block([DB, Acon])
        Bcon = np.array([N1]).T
        Ccon = np.append([1], ZR)
        Dcon = np.array([d1])



    elif struct == 2:
        con1 = ct.tf(ct.tf(controller.num[0][0], controller.den[0][0]))
        con2 = ct.tf(ct.tf(controller.num[0][1], controller.den[0][1]))
        N1, D1 = ct.tfdata(con1)
        N2, D2 = ct.tfdata(con2)

        N1 = N1[0][0]
        D1 = D1[0][0]
        N1 = N1 / D1[0]
        D1 = D1 / D1[0]
        N2 = N2[0][0]
        D2 = D2[0][0]
        N2 = N2 / D2[0]
        D2 = D2/ D2[0]
        print(N1, D1)
        print(N2, D2)

        if len(N1) == len(D1):
            d1 = N1[0]
            N1 = N1 - d1* D1
            N1 = N1[1:]
        else:
            d1 = 0

        if len(N2) == len(D2):
            d2 = N2[0]
            N2 = N2 - d2* D2
            N2 = N2[1:]
        else:
            d2 = 0

        DB = np.array([-D1[1:]])
        DB = DB.T
        size = len(D1)-2
        In_1 = np.eye(size)
        ZR = np.zeros((1,size))
        Acon = np.block([[In_1], [ZR]])
        Acon = np.block([DB, Acon])
        B1 = np.array([N1]).T
        B2 = np.array([N2]).T
        Bcon = np.block([B1,B2])
        Ccon = np.append([1], ZR)
        Dcon = np.block([d1, d2])


    Ad, Bd, Cd, Dd, dt = cont2discrete((Acon, Bcon, Ccon, Dcon), sampling_time, method='bilinear')
    Cve = ct.ss(Ad, Bd, Cd, Dd)
    Ai, Bi, Ci, Di = Cve.A, Cve.B[:, 0], Cve.C, Cve.D[0][0]
    int_system = ct.ss(Ai, Bi, Ci, Di)
    int_system, T = ct.canonical_form(int_system)
    Cve = ct.similarity_transform(Cve, T)
    A = Cve.A
    B = Cve.B
    Cc = Cve.C
    Dc = Cve.D
    order = np.size(A, 0)
    In = np.diag([10000 for i in range(order)])
    L, S, E = ct.dlqr(np.transpose(A), np.transpose(Cc), In, 1)
    L = np.transpose(L)
    Ac = A - L * Cc
    Bc = B - L * Dc
    if struct == 1:
        Bc = np.block([Bc, -Bc])
        Dc = np.array([[Dc[0][0], -Dc[0][0]]])
    A_hex = matrix2hex(Ac)
    B_hex = matrix2hex(Bc)
    C_hex = matrix2hex(Cc)
    D_hex = matrix2hex(Dc)
    L_hex = matrix2hex(L)
    order_hex = long2hex(order)
    type_control_hex = long2hex(type_control)
    message = json.dumps({"order": order_hex,
                          "A": A_hex,
                          "B": B_hex,
                          "C": C_hex,
                          "D": D_hex,
                          "L": L_hex,
                          "typeControl": type_control_hex,
                          })

    system.connect()
    system.publish(topic_pub, message)
    system.disconnect()
    return


def profile_closed(system, timevalues = [0, 50, 100 , 150], refvalues = [40, 50, 60, 70]):
    def profile_message(system, userdata, message):
        # This is the callback for receiving messages from the plant
        q.put(message)

    # reading the configuration parameters from the code's field in the plant

    topic_pub = system.codes["USER_SYS_PROFILE_CLOSED"]
    topic_sub = system.codes["SYS_USER_SIGNALS_CLOSED"]
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]


    # setting the parameters of the step response for sending to ESP32

    int_timevalues = [round(p/sampling_time) for p in timevalues]
    if int_timevalues[0] != 0:
        int_timevalues.insert(0, int_timevalues[0]-1)
        int_timevalues.insert(0,0)
        refvalues.insert(0,0)
        refvalues.insert(0,0)


    int_timevalues_hex = time2hex(int_timevalues)
    refvalues_hex = signal2hex(refvalues)
    points = len(int_timevalues)
    points_hex = long2hex(points)

    # user's command for obtaining the profile
    # al values are transmitted in hexadecimal
    min_val = np.min(refvalues)
    max_val = np.max(refvalues)
    min_val_hex = float2hex(min_val)
    max_val_hex = float2hex(max_val)

    message = json.dumps({"timevalues":  int_timevalues_hex,
                          "refvalues":   refvalues_hex,
                          "points":      points_hex,
                          "min_val":     min_val_hex,
                          "max_val":     max_val_hex,
                          })

    # setting the callback fro receiving data from the ESP32 for obtaining the profile response
    system.client.on_message = profile_message

    # connecting the system
    system.connect()

    # subscribing to topic published by ESP32
    system.subscribe(topic_sub)

    # command sent to ESP32 for obtaining the profile response
    system.publish(topic_pub, message)

    # setting the total of points and the total of frames
    total_points = int_timevalues[-1]

    q = Queue()
    y = []
    r = []
    u = []
    t = []


    # Setting the graphics configuration for visualizing the experiment
    fig, (ay, au) = plt.subplots(nrows=2, ncols=1, width_ratios = [1], height_ratios= [4,1], figsize=(16, 9))
    fig.set_facecolor('#b7c4c8f0')

    # settings for the upper axes, depicting the model and speed data
    ay.set_title(f'Profile response experiment with a duration of {timevalues[-1]:0.2f} seconds and {len(timevalues):d} edges')
    ay.set_ylabel('')
    ay.grid(True);
    ay.grid(color='#806600ff', linestyle='--', linewidth=0.25)
    ay.set_facecolor('#f4eed7ff')
    ay.set_xlim(0, timevalues[-1])

    #Setting the limits of figure
    ylimits = [np.min(refvalues) - 5, np.max(refvalues) + 5]
    ay.set_ylim(ylimits[0], ylimits[1])

    au.set_facecolor('#d7f4e3ff')
    au.set_ylim(0, 100)
    au.set_xlim(0, timevalues[-1])
    au.grid(color='#008066ff', linestyle='--', linewidth=0.25)


    line_r, = ay.plot(t, r, color="#008066ff", linewidth=1.25)
    line_y, = ay.plot(t, y, color="#ff0066ff")
    line_u, = au.plot(t, u, color="#0066ffff")



    box = dict(boxstyle='round,pad=0.5', facecolor='white', edgecolor='white', alpha=0.75)
    # y_txt = ay.text( t0 + t1- 3, ylimits[0] + 1 , 'Temperature:\n ', fontsize=15, color="#ff6680",ha='right', va='bottom', bbox=box)
    # u_txt = au.text(t0 + t1 - 3 , 15, f'Input:', fontsize=15, color="#0066ffB0",ha='right', va='bottom', bbox=box)


    exp = []
    n = -1
    sync = False

    while n < total_points:
        try:
            message = q.get(True, 20 * sampling_time)
        except:
            raise TimeoutError("The connection has been lost. Please try again")

        decoded_message = str(message.payload.decode("utf-8"))
        msg_dict = json.loads(decoded_message)
        n_hex = str(msg_dict["np"])
        n = hex2long(n_hex)
        print(n)
        if n == 0:
            sync = True
        if sync == True:
            t_curr = n * sampling_time
            t.append(t_curr)
            y_curr = hex2float(msg_dict["y"])
            y.append(y_curr)
            r_curr = hex2float(msg_dict["r"])
            r.append(r_curr)
            u_curr = hex2float(msg_dict["u"])
            u.append(u_curr)
            line_r.set_data(t, r)
            line_y.set_data(t, y)
            line_u.set_data(t, u)
            ay.legend([line_r, line_y], [f'$r(t):$ {r_curr:0.2f}', f'$y(t):$ {y_curr: 0.3f}$~^oC$'], fontsize=16, loc="upper left")
            au.legend([line_u], [f'$u(t):$ {u_curr: 0.1f}'], fontsize=16)
            exp.append([t_curr, r_curr, y_curr, u_curr])
            plt.draw()
            plt.pause(sampling_time)
    ay.legend([line_r, line_y], ['$r(t)$ (reference)', '$y_R(t)$ (real output)'], fontsize=16,
              loc="upper left")
    au.legend([line_u], ['$u(t)$ (control signal)'], fontsize=14)



    np.savetxt(PATH_DEFAULT + "Thermal_profile_closed_exp.csv", exp, delimiter=",", fmt="%0.8f",
               comments="", header='t,r,y,u')
    np.savetxt(PATH_DATA + "Thermal_profile_closed_exp.csv", exp, delimiter=",", fmt="%0.8f",
               comments="", header='t,r,y,u')
    plt.show()
    system.disconnect()
    return t, r, y, u










if __name__ == "__main__":
    plant = ThermalSystemIoT()
    set_pid(plant, kp=16.796, ki=2, kd=16.441, N=27.38, beta=0.5)
    t,  r, y, u = step_closed(plant, 40, 60, 30, 30)
    #signal = [30, 40, 50, 60, 70, 80]
    # set_pid(plant, kp = 16.796, ki = 5, kd = 16.441, N = 27.38, beta = 1)
    #step_open(plant, op_point=45, amplitude=5, high_time=200, stab_time=150, uee_time=20)
    #pbrs_open(plant, op_point=45, peak_amp=5, stab_time=150, uee_time=20, divider=35)
    #stair_closed(plant, signal, 10)





# Required libraries
import paho.mqtt.client as mqtt
import control as ct
import struct
from queue import Queue
from math import ceil
import json
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
from scipy.signal import cont2discrete

#


# parameters of communication


BROKER = "192.168.0.3"
PORT = 1883
USER = "hpdesktop"
PASSWORD = "hpdesktop"

#topics for subscribing

PLANT_NUMBER = "1234"
codes ={"SYS_USER_SIGNALS_CLOSED"  : "/thermal/thermal_" + PLANT_NUMBER + "/user/sig_closed",
        "SYS_USER_SIGNALS_OPEN"  : "/thermal/thermal_" + PLANT_NUMBER + "/user/sig_open",
        "USER_SYS_SET_REF"  : "/thermal/user/thermal_" + PLANT_NUMBER + "/set_ref",
        "USER_SYS_SET_PID"  : "/thermal/user/thermal_" + PLANT_NUMBER  + "/set_pid",
        "USER_SYS_STEP_CLOSED": "/thermal/user/thermal_" + PLANT_NUMBER +"/step_closed",
        "USER_SYS_STAIRS_CLOSED": "/thermal/user/thermal_" + PLANT_NUMBER + "/stairs_closed",
        "USER_SYS_PRBS_OPEN": "/thermal/user/thermal_" + PLANT_NUMBER + "/prbs_open",
        "USER_SYS_STEP_OPEN": "/thermal/user/thermal_" + PLANT_NUMBER + "/step_open",
        "USER_SYS_SET_GENCON": "/thermal/user/thermal_" + PLANT_NUMBER + "/set_gencon",
        "USER_SYS_PROFILE_CLOSED": "/thermal/user/thermal_" + PLANT_NUMBER + "/prof_closed",
        "THERMAL_SAMPLING_TIME" : 0.8
        }


PATH_DEFAULT = r"./experiment_files/"
PATH_DATA = str(Path(__file__).parent) + r"/datafiles/"
Path(PATH_DEFAULT).mkdir(exist_ok=True)
class ThermalSystemIoT:

    def __init__(self, broker_address = BROKER, port= PORT, user=USER, password=PASSWORD,  client_id="", clean_session=True):
        self.client = mqtt.Client()
        self.broker_address = broker_address
        self.port = port
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.on_message = self.on_message
        self.client.on_subscribe = self.on_subscribe
        self.client.on_publish = self.on_publish
        self.codes = codes


    def on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            print("Connected successfully to MQTT Broker!")
        else:
            print("Failed to connect, return code %d\n", rc)

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            print("Unexpected disconnection.")

    def on_message(self, client, userdata, message):
        print(f"Received  '{message.payload.decode()}'")

    def on_subscribe(self, client, userdata, mid, granted_qos):
        print("Subscribed: ", mid, " ", granted_qos)

    def on_publish(self, client, userdata, mid):
        print("Message Published: ", mid)

    def connect(self):
        self.client.username_pw_set(USER, PASSWORD)
        self.client.connect(self.broker_address, self.port)
        self.client.loop_start()

    def disconnect(self):
        self.client.loop_stop()
        self.client.disconnect()

    def subscribe(self, topic, qos=2):
        self.client.subscribe(topic, qos)

    def publish(self, topic, message, qos=2):
        self.client.publish(topic, message, qos)



    def transfer_function(self, temperature=50):
        Kp = -0.0025901782151786 * temperature + 0.987094648761147
        Tao = -0.0973494029141449 * temperature + 66.5927276606595
        delay = -0.00446863636363636 * temperature + 3.57201818181818
        uN = 0.9719 * temperature - 24.7355
        G = ct.TransferFunction(Kp, [Tao, 1])
        return G, delay, uN




from .controlsys import  ThermalSystemIoT
from .identsys import get_models_prbs, get_static_model, prbs_open, step_open
from .controlsys import step_closed, set_controller, set_reference, set_pid, stairs_closed, profile_closed


import matplotlib.pyplot as plt
import numpy as np
from scipy.integrate import odeint
from scipy.optimize import minimize
from scipy.interpolate import interp1d
from scipy.stats import linregress
import control as ct
from .thermalsys import ThermalSystemIoT, PATH_DATA, PATH_DEFAULT
from .controlsys import  long2hex, float2hex, hex2long, set_pid, hex2float
import json
from math import ceil
from queue import Queue
from pathlib import Path
import csv
from pathlib import Path


# import matplotlib
# matplotlib.use("TkAgg", force=True)






def read_csv_file3():
    with open(PATH_DATA + 'Thermal_prbs_open_exp.csv', newline='') as file:
        reader = csv.reader(file)
        # Iterate over each row in the CSV file
        num_line = 0
        t = []
        u = []
        y = []
        for row in reader:
            if num_line != 0:
               t.append(float(row[0]))
               u.append(float(row[1]))
               y.append(float(row[2]))
            num_line += 1
        return t, u, y




def step_closed_staticgain(system,r0=0,r1=40,t0=0,t1=40):
    def step_message(system, userdata, message):
        q.put(message)


    low_val = r0
    high_val = r1
    low_time = t0
    high_time = t1

    topic_pub = system.codes["USER_SYS_STEP_CLOSED"]
    topic_sub = system.codes["SYS_USER_SIGNALS_CLOSED"]
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]
    points_high = round(high_time / sampling_time)
    points_low = round(low_time / sampling_time)
    points_low_hex = long2hex(points_low)
    points_high_hex = long2hex(points_high)
    low_val_hex = float2hex(low_val)
    high_val_hex = float2hex(high_val)
    message = json.dumps({"low_val": low_val_hex,
                          "high_val": high_val_hex,
                          "points_low": points_low_hex,
                          "points_high": points_high_hex,
                          })

    system.client.on_message = step_message
    system.connect()
    system.subscribe(topic_sub)
    system.publish(topic_pub, message)
    q = Queue()
    n = 0
    y = []
    u = []
    t = []
    ax = plt.gca()
    line_y, = ax.plot(t, y, linestyle = 'solid', color="#0044AA60", linewidth=1)
    line_u, = ax.plot(t, u, linestyle='solid', color="#ff000060", linewidth=1)
    ax.legend([line_y, line_u], [r'Temperature $(~^o C)$', r'Power input ($\%$ of 2.475W)'], fontsize=16 , loc = "lower right")
    points = points_high + points_low
    n = 0
    sync = False
    while n <= points:
        try:
            message = q.get(True, 20 * sampling_time)
        except:
            raise TimeoutError("The connection has been lost. Please try again")

        decoded_message = str(message.payload.decode("utf-8"))
        msg_dict = json.loads(decoded_message)
        n_hex = str(msg_dict["np"])
        n = hex2long(n_hex)
        if n==0:
            sync = True
        if sync:
            t_curr = n * sampling_time
            t.append(t_curr)
            y_curr = hex2float(msg_dict["y"])
            y.append(y_curr)
            u_curr = hex2float(msg_dict["u"])
            u.append(u_curr)
            line_y.set_data(t, y)
            line_u.set_data(t, u)
            plt.draw()
            plt.pause(sampling_time)
    line_y.set_data([], [])
    line_u.set_data([], [])
    system.disconnect()
    return u, y




def get_static_model(system, step = 5):

    # This is the configuration for the figure displayed while acquiring data
    yee = []
    uee = []
    fig, ax = plt.subplots(figsize=(16, 9))
    fig.set_facecolor('#b7c8be')
    ax.set_title('Static gain response experiment for UNThermalSystem')
    ax.set_xlabel(r'Percent of Power ($\%$ of 2.475W) / Seconds for the current experiment')
    ax.set_ylabel(r'Steady state temperature (C)')
    ax.set_facecolor('#f4eed7')
    ax.set_xticks([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100])
    ax.set_yticks([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100])
    ax.grid(color='#1a1a1a40', linestyle='--', linewidth=0.25)
    line_exp, = ax.plot(uee, yee, color="#00aa00", linewidth=1, marker=r"$\circ$",markeredgewidth=0.1)
    ax.set_xlim(0, 100)
    ax.set_ylim(0, 100)
    exp = []
    y_test = np.arange(30,100, step)


    for yi in y_test:
        u, y = step_closed_staticgain(system, r0=0, r1=yi, t0=0, t1=60)
        yf = np.mean(y[-15:])
        uf = np.mean(u[-15:])
        exp.append([uf, yf])
        yee.append(yf)
        uee.append(uf)
        line_exp.set_data(uee, yee)
        plt.draw()
        plt.pause(0.1)

    np.savetxt(PATH_DEFAULT + "Thermal_static_gain_response.csv", exp, delimiter=",", fmt="%0.8f", comments="",
               header='u,t')
    np.savetxt(PATH_DATA + "Thermal_static_gain_response.csv", exp, delimiter=",", fmt="%0.8f", comments="",
               header='u,t')

    res = linregress(uee, yee, alternative='greater')
    m = res.slope
    b = res.intercept
    ymod = m * np.array(uee) + b
    line_mod, = ax.plot(uee, ymod, color="#0088aaff", linewidth=1.5)
    stringmodel = r"Model:  $T_{ee}= m\,u_{ee} + b=$" + f"{m:0.2f}" + r"$\,u_{ee}+$" + f"{b:0.2f}"
    ax.legend([line_exp, line_mod], ['Data', stringmodel], fontsize=16)
    np.savetxt(PATH + "Thermal_static_gain_model.csv", [m,b], delimiter=",", fmt="%0.8f", comments="",
               header='m,b')
    plt.show()
    system.disconnect()
    return



def prbs_open(system, op_point=50, peak_amp=4, stab_time=60, uee_time=10, divider=30):
    def pbrs_message(system, userdata, message):
        q.put(message)

    topic_pub = system.codes["USER_SYS_PRBS_OPEN"]
    topic_sub = system.codes["SYS_USER_SIGNALS_OPEN"]
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]
    peak_amp_hex = float2hex(peak_amp)
    op_point_hex = float2hex(op_point)
    stab_points = ceil(stab_time / sampling_time)
    uee_points = ceil(uee_time / sampling_time)
    stab_points_hex = long2hex(stab_points)
    uee_points_hex = long2hex(uee_points)
    divider_hex = long2hex(divider)
    points = divider * 63 + stab_points + uee_points
    message = json.dumps({"peak_amp": peak_amp_hex,
                          "op_point": op_point_hex,
                          "stab_points": stab_points_hex,
                          "uee_points": uee_points_hex,
                          "divider": divider_hex
                          })
    system.client.on_message = pbrs_message
    system.connect()
    system.subscribe(topic_sub)
    system.publish(topic_pub, message)
    q = Queue()

    y = []
    u = []
    t = []
    yt = []
    ut = []
    tt = []
    exp = []
    m = 1.2341015052212259
    b = 24.750915901094388
    uf_est = (op_point - b) / m
    percent = 0.2
    ymax = op_point + m*peak_amp + 2
    ymin = op_point - m*peak_amp - 2
    umax = uf_est + (1 + percent) * peak_amp
    umin =  np.min([0, uf_est - (1 + percent) * peak_amp - 5.0])
    fig, (yax, uax) = plt.subplots(nrows=2, ncols=1, width_ratios=[1], height_ratios=[2, 1], figsize=(16, 9))
    fig.set_facecolor('#b7c4c8f0')
    yax.set_title(f'PRBS identification with {points:d} samples and a duration of {points * sampling_time: 0.2f} seconds')
    yax.set_ylabel(r'Temperature ($~^oC$)')
    yax.grid(True);
    yax.grid(color='#1a1a1a40', linestyle='--', linewidth=0.25)
    yax.set_facecolor('#f4eed7')
    yax.set_xlim(0, sampling_time * points)
    yax.set_ylim(ymin, ymax)
    yax.set_xlabel('Time (s)')
    uax.set_xlabel('Time (s)')
    uax.set_ylabel(r'Power Input ($\%$)')
    uax.grid(True);
    uax.set_facecolor('#d7f4ee')
    uax.grid(color='#1a1a1a40', linestyle='--', linewidth=0.25)
    uax.set_xlim(0, sampling_time * points)
    uax.set_ylim(umin, umax)

    line_y, = yax.plot(t, y, color="#ff6680")
    line_u, = uax.plot(t, u, color="#00d4aa")
    line_yt, = yax.plot(t, yt, color="#d40055")
    line_ut, = uax.plot(t, ut, color="#338000")
    box = dict(boxstyle='round,pad=0.5', facecolor='white', edgecolor='white', alpha=0.75)
    txt1 = yax.text(25, ymin + 1, 'Current Temperature:', fontsize=15, color="#ff6680",ha='left', va='bottom', bbox=box)
    txt2 = uax.text(25, 5, f'Current Input:', fontsize=15, color="#00d4aa",ha='left', va='bottom', bbox=box)
    n = -1
    sync = False
    while n < points:
        try:
            message = q.get(True, 20 * sampling_time)
        except:
            raise TimeoutError("The connection has been lost. Please try again")

        decoded_message = str(message.payload.decode("utf-8"))
        msg_dict = json.loads(decoded_message)
        n_hex = str(msg_dict["np"])
        n = hex2long(n_hex)
        if n == 0:
            sync = True

        if sync == True:
            if n <= stab_points + uee_points:
                t_curr = n*sampling_time
                t.append(t_curr)
                y_curr = hex2float(msg_dict["y"])
                y.append(y_curr)
                u_curr = hex2float(msg_dict["u"])
                u.append(u_curr)
                line_y.set_data(t, y)
                line_u.set_data(t, u)
                txt1.set_text( f'Current Temperature: {y_curr: 0.2f}$~^oC$')
                txt2.set_text( f'Current Input: {u_curr:0.2f}% ({0.02475*u_curr:0.2f} W)')
                if n == stab_points + uee_points:
                    tt.append(t_curr)
                    yt.append(y_curr)
                    ut.append(u_curr)
                    exp.append([0, u_curr, y_curr])
            else:
                tt_curr = n * sampling_time
                if n == stab_points + uee_points + 1:
                    t0 = t_curr
                tt.append(tt_curr)
                yt_curr = hex2float(msg_dict["y"])
                yt.append(yt_curr)
                ut_curr = hex2float(msg_dict["u"])
                ut.append(ut_curr)
                line_yt.set_data(tt, yt)
                line_ut.set_data(tt, ut)
                line_y.set_data(t, y)
                line_u.set_data(t, u)
                txt1.set_color("#d40055")
                txt2.set_color("#338000")
                txt1.set_text( f'Current Temperature: {yt_curr: 0.2f}$~^oC$')
                txt2.set_text( f'Current Input: {ut_curr:0.2f}% ({0.02475*ut_curr:0.2f} W)')
                exp.append([tt_curr-t0, ut_curr, yt_curr])

            plt.draw()
            plt.pause(0.1)

    np.savetxt(PATH_DEFAULT + "prbs_open_exp.csv", exp, delimiter=",", fmt="%0.8f", comments="", header='t,u,y')
    np.savetxt(PATH_DATA + "prbs_open_exp.csv", exp, delimiter=",", fmt="%0.8f", comments="", header='t,u,y')
    system.disconnect()
    return tt, ut, yt



def get_models_prbs(system, yop = 50, peak_amp= 4, usefile = True):

    norm = np.linalg.norm

    def simulate_fo_model(x):
        # this function simulates the model
        alpha, tau = x
        s = ct.TransferFunction.s
        G = alpha / (tau*s + 1)
        tsim, ysim = ct.forced_response(G, t, um)
        return G, ysim

    # def simulate_so_model(x):
    #     # this function simulates the model
    #     alpha, tau1, tau2 = x
    #     s = ct.TransferFunction.s
    #     G = alpha / ((tau1*s + 1)*(tau2*s+1))
    #     tsim, ysim = ct.forced_response(G, t, um)
    #     return G, ysim

    def simulate_fotd_model(x):
        # this function simulates the model
        alpha, tau1, tau2 = x
        um_delay = um_interp(np.array(t) - tau2)
        s = ct.TransferFunction.s
        G = alpha / ( tau1*s + 1)
        tsim, ysim = ct.forced_response(G, t, um_delay)
        return G, ysim

    def objective_fo(x):
        # simulate model
        G, ysim = simulate_fo_model(x)
        # return objective
        return norm(ysim - ym)

    def objective_so(x):
        # simulate model
        G, ysim = simulate_so_model(x)
        # return objective
        return norm(ysim - ym)

    def objective_fotd(x):
        # simulate model
        G, ysim = simulate_fotd_model(x)
        # return objective
        return norm(ysim - ym)


    if  (yop >= 100):
         raise ValueError(f"The maximum temperature for this system is 100 degrees celsius")

    if usefile:
        t, u, y = read_csv_file3()
    else:
        t, u, y = prbs_open(system, op_point=yop, peak_amp=peak_amp, stab_time=60, uee_time=10, divider=30)

    m = 1.2341015052212259
    ymean = np.mean(y)
    um = np.array(u) - u[0]
    ym = np.array(y) - ymean

    um_interp = interp1d(t, um, fill_value = (0,0 ), bounds_error=False)

    """Now we obtain the initial parameters for the model
                        alpha
              G1(s) = -----------
                      (tau1*s +1)
     and 
                      
                         alpha
              G2(s) = -----------  exp(-tau2 s)
                      (tau1*s +1)
                      
                      
    """
    alpha_0 = 1.23
    tau1_0 = 60
    tau2_0 = 2
    x02 = [alpha_0, tau1_0, tau2_0]
    x01 = [alpha_0, tau1_0]
      # # These are the bounds for alpha, tau1 and tau2

    bounds2 = [(0.8, 1.5), (1 , 150), (1, 20)]
    bounds1 = bounds2[0:2]
    #
    # Now we run the optimization algorithm
    print(f'Starting optimization for first order model...\n\t Initial cost function: {objective_fo(x01):.2f}' )
    solution = minimize(objective_fo, x0=x01, bounds= bounds1)
    xmin = solution.x
    alpha, tau = xmin
    fmin = objective_fo(xmin)
    print(f'\t Final cost function: {fmin:.2f}' )
    print(f'alpha={alpha:.2f} \t tau1={tau:.3f}')
    #
    # We compare the experimental data with the simulation model
    G1, ysim1 = simulate_fo_model(xmin)
    r1 = 100*(1 - norm(ym - ysim1) / norm(ym))
    #
    print(f'\n\nStarting optimization for second order model...\n\t Initial cost function: {objective_fotd(x02):.2f}' )
    solution = minimize(objective_fotd, x0=x02, bounds= bounds2)
    xmin = solution.x
    alpha2, tau1, tau2 = xmin
    fmin = objective_fotd(xmin)
    print(f'\t Final cost function: {fmin:.2f}' )
    print(f'alpha={alpha2:.2f} \t tau1={tau1:.3f} \t tau2={tau2:.3f}')

    # We compare the experimental data with the simulation model
    G2, ysim2 = simulate_fotd_model(xmin)
    r2 = 100*(1 - norm(ym - ysim2) / norm(y - np.mean(y)))

    # we calculate the step response from the model
    # now we compare the model with the experimental data
    fig, (ay, au) = plt.subplots(nrows=2, ncols=1, width_ratios=[1], height_ratios=[5, 1], figsize=(16, 9))
    fig.set_facecolor('#b7c4c8f0')

    # settings for the upper axes, depicting the model and speed data
    #ay.set_title('Data and estimated second order model for UNDCMotor')
    ay.set_ylabel(r'Celsius degrees ($^oC$)')
    ay.grid(True)
    ay.grid(color='#1a1a1a40', linestyle='--', linewidth=0.125)
    ay.set_facecolor('#f4eed7')
    ay.set_xlim(0, t[-1])
    ay.set_xlabel('Time (seconds)')

    # settings for the lower axes, depicting the input
    au.set_xlim(0, t[-1])
    au.set_facecolor('#d7f4ee')
    au.grid(color='#1a1a1a40', linestyle='--', linewidth=0.125)

    au.set_xlabel('Time (seconds)')
    au.set_ylabel('Power (%)')

    line_exp, = ay.plot(t, ym + ymean, color="#0088aaAF", linewidth=1.5, linestyle=(0, (1, 1)))
    line_model1, = ay.plot(t, ysim1 + ymean, color="#00AA44ff", linewidth=1.5, )
    line_model2, = ay.plot(t, ysim2 + ymean, color="#d45500ff", linewidth=1.5, )


    #ay.plot(timestep + tau, ya + 0.63212 * delta_y, color="#ff0066", linewidth=1.5, marker=".", markersize=13)
    line_u, = au.plot(t, u, color="#0066ffff")
    #line_ud, = au.plot(t, um_interp(np.array(t)-20) , color="#0000ff")
    modelstr1 = r"FO Model:     $G_1(s) = \frac{%0.2f }{%0.2f\,s+1}$  ($FIT = %0.1f$" % (alpha, tau, r1) + "%)"
    modelstr2 = r"FOTD Model: $G_2(s) = \frac{%0.2f  }{%0.2f\,s+1}\,e^{-%0.2f\,s}$  ($FIT = %0.1f$"%(alpha2, tau1, tau2, r2) + "%)"
    ay.set_title("Comparison of FO and FOTD models estimated with a PRBS signal at the operation point $y_{OP}=%0.1f^oC$"%yop)
    ay.legend([line_exp, line_model1, line_model2], ['Data', modelstr1, modelstr2],
              fontsize=15, loc = 'lower left',framealpha=0.95)
    au.legend([line_u], ['PRBS Input'], fontsize=14)
    # PATH1 = r'/home/leonardo/sharefolder/ProyectoSabatico/Reporte/figures/'
    # plt.savefig(PATH1 + "Thermal_pbrs.svg", format="svg", bbox_inches="tight")

    fo_model = [[alpha, tau]]
    so_model = [[alpha2, tau1, tau2]]
    np.savetxt(PATH_DEFAULT + "Thermal_fo_model_pbrs.csv", fo_model, delimiter=",",
               fmt="%0.8f", comments="", header='alpha, tau')

    np.savetxt(PATH_DEFAULT + "Thermal_fotd_model_pbrs.csv", so_model, delimiter=",",
               fmt="%0.8f", comments="", header='alpha2, tau1, tau2')
    np.savetxt(PATH_DATA + "Thermal_fo_model_pbrs.csv", fo_model, delimiter=",",
               fmt="%0.8f", comments="", header='alpha, tau')

    np.savetxt(PATH_DATA + "Thermal_fotd_model_pbrs.csv", so_model, delimiter=",",
               fmt="%0.8f", comments="", header='alpha2, tau1, tau2')

    system.disconnect()
    plt.show()
    return #G1, G2

def step_open(system, op_point=50, amplitude=5, high_time=200, stab_time=150, uee_time=20):
    def step_message(system, userdata, message):
        q.put(message)

    topic_pub = system.codes["USER_SYS_STEP_OPEN"]
    topic_sub = system.codes["SYS_USER_SIGNALS_OPEN"]
    sampling_time = system.codes["THERMAL_SAMPLING_TIME"]
    amp_hex = float2hex(amplitude)
    points_high = ceil(high_time / sampling_time)
    points_high_hex = long2hex(points_high)
    op_point_hex = float2hex(op_point)
    stab_points = ceil(stab_time / sampling_time)
    uee_points = ceil(uee_time / sampling_time)
    stab_points_hex = long2hex(stab_points)
    uee_points_hex = long2hex(uee_points)
    message = json.dumps({"amplitude": amp_hex,
                          "op_point": op_point_hex,
                          "stab_points": stab_points_hex,
                          "uee_points": uee_points_hex,
                          "points_high": points_high_hex,
                          })
    system.client.on_message = step_message
    system.connect()
    system.subscribe(topic_sub)
    system.publish(topic_pub, message)
    q = Queue()
    y = []
    u = []
    t = []
    yt = []
    ut = []
    tt = []
    exp = []
    m = 1.2341015052212259
    b = 24.750915901094388
    uf_est = (op_point - b) / m
    percent = 0.2
    ymax = op_point + m* amplitude + 2
    ymin = op_point - 2
    umax = uf_est + (1 + percent) * amplitude
    umin =  np.min([0, uf_est - (1 + percent) * amplitude])
    points = stab_points + uee_points + points_high
    fig, (yax, uax) = plt.subplots(nrows=2, ncols=1, width_ratios=[1], height_ratios=[2, 1], figsize=(16, 9))
    fig.set_facecolor('#b7c4c8f0')
    yax.set_title(f'PRBS identification with {points:d} samples and a duration of {points * sampling_time: 0.2f} seconds')
    yax.set_ylabel(r'Temperature ($~^oC$)')
    yax.grid(True);
    yax.grid(color='#1a1a1a40', linestyle='--', linewidth=0.25)
    yax.set_facecolor('#f4eed7')
    yax.set_xlim(0, sampling_time * points)
    yax.set_ylim(ymin, ymax)
    yax.set_xlabel('Time (s)')
    uax.set_xlabel('Time (s)')
    uax.set_ylabel(r'Power Input ($\%$)')
    uax.grid(True);
    uax.set_facecolor('#d7f4ee')
    uax.grid(color='#1a1a1a40', linestyle='--', linewidth=0.25)
    uax.set_xlim(0, sampling_time * points)
    uax.set_ylim(umin, umax)

    line_y, = yax.plot(t, y, color="#ff6680")
    line_u, = uax.plot(t, u, color="#00d4aa")
    line_yt, = yax.plot(t, yt, color="#d40055")
    line_ut, = uax.plot(t, ut, color="#338000")
    box = dict(boxstyle='round,pad=0.5', facecolor='white', edgecolor='white', alpha=0.75)
    txt1 = yax.text(25, ymin + 1, 'Current Temperature:', fontsize=15, color="#ff6680",ha='left', va='bottom', bbox=box)
    txt2 = uax.text(25, 5, f'Current Input:', fontsize=15, color="#00d4aa",ha='left', va='bottom', bbox=box)
    n = -1
    sync = False
    while n < points:
        try:
            message = q.get(True, 20 * sampling_time)
        except:
            raise TimeoutError("The connection has been lost. Please try again")

        decoded_message = str(message.payload.decode("utf-8"))
        msg_dict = json.loads(decoded_message)
        n_hex = str(msg_dict["np"])
        n = hex2long(n_hex)
        if n == 0:
            sync = True

        if sync == True:
            if n <= stab_points + uee_points:
                t_curr = n*sampling_time
                t.append(t_curr)
                y_curr = hex2float(msg_dict["y"])
                y.append(y_curr)
                u_curr = hex2float(msg_dict["u"])
                u.append(u_curr)
                line_y.set_data(t, y)
                line_u.set_data(t, u)
                txt1.set_text( f'Current Temperature: {y_curr: 0.2f}$~^oC$')
                txt2.set_text( f'Current Input: {u_curr:0.2f}% ({0.02475*u_curr:0.2f} W)')
                if n == stab_points + uee_points:
                    tt.append(t_curr)
                    yt.append(y_curr)
                    ut.append(u_curr)
                    exp.append([0, u_curr, y_curr])
            else:
                tt_curr = n * sampling_time
                if n == stab_points + uee_points + 1:
                    t0 = t_curr
                tt.append(tt_curr)
                yt_curr = hex2float(msg_dict["y"])
                yt.append(yt_curr)
                ut_curr = hex2float(msg_dict["u"])
                ut.append(ut_curr)
                line_yt.set_data(tt, yt)
                line_ut.set_data(tt, ut)
                line_y.set_data(t, y)
                line_u.set_data(t, u)
                txt1.set_color("#d40055")
                txt2.set_color("#338000")
                txt1.set_text( f'Current Temperature: {yt_curr: 0.2f}$~^oC$')
                txt2.set_text( f'Current Input: {ut_curr:0.2f}% ({0.02475*ut_curr:0.2f} W)')
                exp.append([tt_curr-t0, ut_curr, yt_curr])

            plt.draw()
            plt.pause(0.1)
    # PATH1 = r'/home/leonardo/sharefolder/ProyectoSabatico/Reporte/figures/'
    # plt.savefig(PATH1 + "Thermal_pbrs.svg", format="svg", bbox_inches="tight")
    np.savetxt(PATH_DEFAULT, exp, delimiter=",",fmt="%0.8f", comments="", header='t,u,y')
    np.savetxt(PATH_DATA, exp, delimiter=",",fmt="%0.8f", comments="", header='t,u,y')
    system.disconnect()
    plt.show()
    return tt, ut, yt



def read_fo_model():
    with open(PATH_DATA + 'Thermal_fo_model_pbrs.csv', newline='') as file:
        reader = csv.reader(file)
        # Iterate over each row in the CSV file
        num_line = 0
        for row in reader:
            if num_line != 0:
               alpha = float(row[0])
               tau = float(row[1])
            num_line += 1
    b = alpha / tau
    a = 1 / tau
    G = ct.tf(b, [1, a])
    return G




if __name__ == "__main__":
    plant = ThermalSystemIoT()
    get_models_prbs(plant, yop=55, peak_amp=4, usefile=False)



